import os
import sys

from typing import Iterable, Callable

from .hash_observer import HashObserver
from .constants import EXCLUDE_RECOMMENDED


__all__ = [
    'observe',
]

TIP = """Observing path "{path}" triggering `{function}()` excluding:
"{excluded}".
"""


def observe(callback: Callable, path: str = '.', exclude: Iterable[str] = None):
    """Observe directory and trigger callback on change detection.

    :param callback: Target function to be called
    :param path: The observable root path
    :param exclude: Excluded files (default is tree_guardian.EXCLUDE_RECOMMENDED)
    """

    exclude = exclude or EXCLUDE_RECOMMENDED

    sys.stdout.write(TIP.format(
        path=os.path.abspath(path),
        function=callback.__name__,
        excluded='"; "'.join(exclude)
    ))
    sys.stdout.flush()

    observer = HashObserver(path=path, exclude=exclude)
    observer.observe(callback=callback)



import os
import re
import time
import hashlib

from typing import Iterable, Callable


class HashObserver(object):

    def __init__(self, path: str = '.', exclude: Iterable[str] = None):
        self.path = path
        self.excluded = set(exclude or ('.idea', 'venv', '.git'))

        self._stopped = False

    def filter_excluded(self, names: Iterable) -> Iterable:
        def regex_excluded(item):
            for exclude in self.excluded:
                if '*' in exclude and re.match(exclude.replace('*', '(.*)'), item):
                    return True
            return False

        res = []
        for name in names:
            if name in self.excluded or regex_excluded(name):
                continue
            res.append(name)

        return res

    def walk(self, abspath: str):
        if not os.path.isdir(abspath):
            return [abspath, ]

        res = []
        items = self.filter_excluded(os.listdir(abspath))
        for item in items:
            res.extend(self.walk(os.path.join(abspath, item)))

        return res

    @property
    def hash(self):
        path_hash = hashlib.md5()
        for filepath in self.walk(os.path.abspath(self.path)):
            try:
                with open(filepath, 'rb') as file:
                    filehash = hashlib.md5(file.read()).hexdigest()
                    path_hash.update(filehash.encode())
            except Exception as e:
                print(f'Error {e}')

        return path_hash.hexdigest()

    def observe(self, callback: Callable, *args, **kwargs):
        hsh = self.hash
        self._stopped = False
        while not self._stopped:
            time.sleep(0.1)
            if self.hash == hsh:
                continue
            hsh = self.hash
            callback(*args, **kwargs)

    def stop(self):
        self._stopped = True



import os

from .utils import get_gitignored


EXCLUDE_SELF = {os.path.basename(__file__), }

EXCLUDE_DOC = {'*.md', '*.doc', '*.docx', '*.txt', '*.pdf', }

EXCLUDE_CONFIG = {'*.cfg', '*.ini', '*.conf', }
EXCLUDE_YAML = {'*.yml', '*.yaml', }

EXCLUDE_BASH = {'*.sh', }
EXCLUDE_GIT = {'.git', '.gitignore', }
EXCLUDE_DOCKER = {'Dockerfile', '.dockerignore', }
EXCLUDE_VIRTUALENV = {'venv', 'virtualenv'}
EXCLUDE_DEVELOPMENT_TOOLS = {'.idea', }

EXCLUDE_IMAGES = {'*.png', '*.jpg', '*.jpeg', '*.bmp', '*.gif', }
EXCLUDE_AUDIO = {'*.png', '*.jpg', '*.jpeg', '*.bmp', '*.gif', }
EXCLUDE_VIDEO = {'*.mkv', '*.avi', }
EXCLUDE_MEDIA = EXCLUDE_AUDIO | EXCLUDE_IMAGES | EXCLUDE_VIDEO

EXCLUDE_PYCOMPILE = {'*.pyc', '__pycache__', }
EXCLUDE_SETUPTOOLS_FILES = {'build', 'dist', '*.egg-info', }
EXCLUDE_PYTHON_BUILD_FILES = EXCLUDE_PYCOMPILE | EXCLUDE_SETUPTOOLS_FILES

EXCLUDE_GITIGNORED = get_gitignored()

EXCLUDE_RECOMMENDED = EXCLUDE_SELF | EXCLUDE_DOC | EXCLUDE_GIT | EXCLUDE_DEVELOPMENT_TOOLS | EXCLUDE_VIRTUALENV | \
    EXCLUDE_MEDIA | EXCLUDE_PYTHON_BUILD_FILES | EXCLUDE_GITIGNORED
